# Copyright (c) 2012 Cloudera, Inc. All rights reserved.

import logging
import time

from cmf.monitor import schema
from cmf.monitor.abstract_mgmt_monitor import AbstractMgmtMonitor
from cmf.monitor.monitor_properties import ConfParser
from cmf.monitor.base import DebugStatusReport
from cmf.monitor.json_metrics_utils import ParseError
from cmf.monitor.navigator.json_parser import JsonParser
from cmf.monitor.cgroup_metrics_utils import add_cgroup_metrics
from cmf.monitor.proc_metrics_utils import add_proc_metrics
from cmf.monitor.role_update import RoleUpdate

LOG = logging.getLogger('NavigatorMonitor')

_ROLETYPE = 'NAVIGATOR'

# Section and key names in cloudera-monitor.properties
# Since CM 4.5
_SECTION = 'navigator'
_SERVICE_NAME = 'service_name'
_ROLE_NAME = 'role_name'
_HOST = 'host'
_NAVIGATOR_DEBUG_PORT = 'navigator_debug_port'
_LOG_DIR = 'log_dir'
_COLLECT_INTERVAL = 'collect_interval'

class NavigatorMonitor(AbstractMgmtMonitor):

  def __init__(self, daemon, registry, cgroup_root=None, cgroup_name=None):
    AbstractMgmtMonitor.__init__(self,
                                 daemon,
                                 registry,
                                 cgroup_root,
                                 cgroup_name,
                                 _SECTION,
                                 _NAVIGATOR_DEBUG_PORT)
    self._last_conf_path = None
    self._metrics_tuple = (None, None, None)
    self._last_reported_timestamp = None
    self._last_collection_ended = None
    self._last_collection_length = None
    self._role_dir_monitors = [ ]
    self._metric_accessors = daemon.metric_accessors

    # Config data
    self._name = None
    self._service_name = None
    self._log_dir = None
    self._metrics_url = None

  def __str__(self):
    return "NavigatorMonitor for %s" % (self._name,)

  def _reload_config(self, confparse):
    """Returns True/False for success"""
    try:
      # Extraction
      name = confparse.get(_SECTION, _ROLE_NAME)
      service = confparse.get(_SECTION, _SERVICE_NAME)
      log_dir = confparse.get_with_default(_SECTION, _LOG_DIR, None)
      host = confparse.get(_SECTION, _HOST)
      navigator_debug_port = confparse.getint(_SECTION, _NAVIGATOR_DEBUG_PORT)
    except ValueError, ex:
      self.record_error(ex)
      LOG.error("Value error in %s: %s" % (confparse.path, ex))
      return False

    metrics_url = self._compute_metrics_url(host, navigator_debug_port)
    if metrics_url is None:
      return False

    # Assign
    self._name = name
    self._service_name = service
    self._log_dir = log_dir
    self._metrics_url = metrics_url

    # Make monitors for the log directory and the data directories.
    self._role_dir_monitors = [ ]
    if self._log_dir is not None:
      self.create_dir_monitor(self._log_dir, self._role_dir_monitors)

    return True

  def collect(self):
    """Overrides AbstractMonitor. Collect Navigator metrics. Invoked
       asynchronously."""
    start = time.time()

    # Reload config if necessary
    confparse = self.conf_parser
    if self._last_conf_path != confparse.path:
      if not self._reload_config(confparse):
        return
      self._last_conf_path = confparse.path

    # Collect directory info
    for mon in self._role_dir_monitors:
      mon.collect()

    # If our pid value is None, it means that the associated process is not
    # running. If this is the case, we skip metric collection.
    if self._pid is None:
      return

    LOG.debug("Fetching metrics at '%s'" % (self._metrics_url,))
    self._metrics_tuple = self.collect_metrics_from_url(self._metrics_url)

    self._last_collection_ended = time.time()
    self._last_collection_length = self._last_collection_ended - start

  def parse_metrics(self, json):
    return JsonParser(json)

  def get_role_update(self):
    """Returns a RoleUpdate to be sent to the firehose"""
    update = RoleUpdate(name=self._name,
                        roletype=_ROLETYPE,
                        service_name=self._service_name)

    if self.add_metrics_from_webserver(self._metrics_tuple,
                                       self._last_reported_timestamp,
                                       update):
      self._last_reported_timestamp = update.timestamp

    add_proc_metrics(update, self._pid, LOG)
    add_cgroup_metrics(update, self._cgroup_root, self._cgroup_name)

    for mon in self._role_dir_monitors:
      update.add_dir_update(mon.get_role_dir_update())

    self.add_unexpected_exit_metrics(update)

    return update

  def add_webserver_metrics_to_role_update(self, update, metrics):
    if metrics.read_requests is not None:
      update.add_metric(schema.READ_REQUESTS,
                        metrics.read_requests.Count)
      update.add_metric(schema.READ_REQUESTS_MEAN_RATE,
                        metrics.read_requests.MeanRate)
      update.add_metric(schema.READ_REQUESTS_FIVE_MIN_RATE,
                        metrics.read_requests.FiveMinuteRate)
      update.add_metric(schema.READ_REQUESTS_FIFTEEN_MIN_RATE,
                        metrics.read_requests.FifteenMinuteRate)
      update.add_metric(schema.READ_REQUESTS_ONE_MIN_RATE,
                        metrics.read_requests.OneMinuteRate)

    if metrics.insert_requests is not None:
      update.add_metric(schema.WRITE_REQUESTS,
                        metrics.insert_requests.Count)
      update.add_metric(schema.WRITE_REQUESTS_MEAN_RATE,
                        metrics.insert_requests.MeanRate)
      update.add_metric(schema.WRITE_REQUESTS_FIVE_MIN_RATE,
                        metrics.insert_requests.FiveMinuteRate)
      update.add_metric(schema.WRITE_REQUESTS_FIFTEEN_MIN_RATE,
                        metrics.insert_requests.FifteenMinuteRate)
      update.add_metric(schema.WRITE_REQUESTS_ONE_MIN_RATE,
                        metrics.insert_requests.OneMinuteRate)

    if metrics.insertion_rate is not None:
      update.add_metric(schema.PERSISTED_REQUESTS,
                        metrics.insertion_rate.Count)
      update.add_metric(schema.PERSISTED_REQUESTS_MEAN_RATE,
                        metrics.insertion_rate.MeanRate)
      update.add_metric(schema.PERSISTED_REQUESTS_FIVE_MIN_RATE,
                        metrics.insertion_rate.FiveMinuteRate)
      update.add_metric(schema.PERSISTED_REQUESTS_FIFTEEN_MIN_RATE,
                        metrics.insertion_rate.FifteenMinuteRate)
      update.add_metric(schema.PERSISTED_REQUESTS_ONE_MIN_RATE,
                        metrics.insertion_rate.OneMinuteRate)

    if metrics.active_read_requests is not None:
      update.add_metric(schema.ACTIVE_READ_REQUESTS,
                        metrics.active_read_requests.Count)

    if metrics.active_write_requests is not None:
      update.add_metric(schema.ACTIVE_WRITE_REQUESTS,
                        metrics.active_write_requests.Count)

  def get_debug_status_report(self):
    return DebugStatusReport(config_items=self.conf_parser.items(_SECTION),
                             errors=(self.last_error,),
                             last_collection_ended=self._last_collection_ended,
                             last_collection_length=self._last_collection_length)

  def get_subject_type(self):
    return _ROLETYPE
